/*
 * Copyright (c) 2013 Nathaniel Jones
 * Massachusetts Institute of Technology
 */

#include <stdio.h>

#include "ray.h" // required by ambient.h
#include "ambient.h"

#include "optix_radiance.h"
#include "commonStructs.h"

#define AMBIENT_ENTRY      1u

#define AMBIENT_RAY        2u
#define AMBIENT_RECORD_RAY 3u

extern AMBTREE	atrunk;		/* our ambient trunk node */

extern double  avsum;		/* computed ambient value sum (log) */
extern unsigned int  navsum;	/* number of values in avsum */
extern unsigned int  nambvals;	/* total number of indirect values */
//extern unsigned int  nambshare;	/* number of values from file */
//extern unsigned long  ambclock;	/* ambient access clock */
//extern unsigned long  lastsort;	/* time of last value sort */
////extern long  sortintvl = SORT_INTVL;	/* time until next sort */
//extern FILE  *ambinp = NULL;		/* auxiliary file for input */
//extern long  lastpos = -1;		/* last flush position */


void createAmbientSamplingCamera( RTcontext context, const VIEW view );
void gatherAmbientRecords(AMBTREE* at, AmbientRecord** records);
int saveAmbientRecords( AmbientRecord* record );

/* from ambient.c */
extern void avsave(AMBVAL *av);

/* from rpict.c */
//extern double  dstrpix;			/* square pixel distribution */


void createAmbientRecords( RTcontext context, const VIEW view, const int width, const int height )
{
	RTbuffer       ambient_record_buffer;
	AmbientRecord* ambient_record_buffer_data;
	RTvariable     temp_var;

	unsigned int scaled_width, scaled_height, generated_record_count, useful_record_count, i;
	time_t kernel_start_time, kernel_end_time; // Timer in seconds for long jobs
	clock_t kernel_start_clock, kernel_end_clock; // Timer in clock cycles for short jobs

	scaled_width = width / optix_amb_scale;
	scaled_height = height / optix_amb_scale;
	generated_record_count = scaled_width * scaled_height * optix_amb_semgents;

	createAmbientSamplingCamera( context, view );

	/* Create buffer for retrieving ambient records. */
	createCustomBuffer3D( context, RT_BUFFER_OUTPUT, sizeof(AmbientRecord), scaled_width, scaled_height, optix_amb_semgents, &ambient_record_buffer );
	applyContextObject( context, "ambient_record_buffer", ambient_record_buffer );
	applyContextVariable1ui( context, "segments", optix_amb_semgents );

	/* Turn off ambient sampling for this calculation. */
	RT_CHECK_ERROR( rtContextQueryVariable( context, "ambacc", &temp_var ) );
	RT_CHECK_ERROR( rtVariableSet1f( temp_var, 0 ) ); // -aa

	/* Run */
	RT_CHECK_ERROR( rtContextValidate( context ) );
	RT_CHECK_ERROR( rtContextCompile( context ) );
	kernel_start_time = time((time_t *)NULL);
	kernel_start_clock = clock();
	RT_CHECK_ERROR( rtContextLaunch2D( context, AMBIENT_ENTRY, scaled_width, scaled_height ) );
	kernel_end_clock = clock();
	kernel_end_time = time((time_t *)NULL);
	fprintf(stderr, "OptiX ambient kernel time: %u milliseconds.\r\n", (kernel_end_clock - kernel_start_clock) * 1000 / CLOCKS_PER_SEC);
	fprintf(stderr, "OptiX ambient kernel time: %u seconds.\r\n", kernel_end_time - kernel_start_time);

	RT_CHECK_ERROR( rtBufferMap( ambient_record_buffer, (void**)&ambient_record_buffer_data ) );
	RT_CHECK_ERROR( rtBufferUnmap( ambient_record_buffer ) );

	/* Copy the results to allocated memory. */
	//TODO the buffer could go directoy to creating Bvh
	useful_record_count = 0u;
	for (i = 0u; i < generated_record_count; i++) {
		useful_record_count += saveAmbientRecords( &ambient_record_buffer_data[i] );
	}
	fprintf(stderr, "Retrieved %u ambient records from %u queries.\r\n", useful_record_count, generated_record_count);

	/* Turn ambient sampling back on. */
	RT_CHECK_ERROR( rtVariableSet1f( temp_var, ambacc ) ); // -aa
	RT_CHECK_ERROR( rtContextQueryVariable( context, "top_ambient", &temp_var ) );
	RT_CHECK_ERROR( rtContextRemoveVariable( context, temp_var ) );

	/* Update ambient average. */
	RT_CHECK_ERROR( rtContextQueryVariable( context, "avsum", &temp_var ) );
	RT_CHECK_ERROR( rtVariableSet1f( temp_var, avsum ) );
	RT_CHECK_ERROR( rtContextQueryVariable( context, "navsum", &temp_var ) );
	RT_CHECK_ERROR( rtVariableSet1ui( temp_var, navsum ) );

	//RT_CHECK_ERROR( rtContextQueryVariable( context, "ambient_record_buffer", &temp_var ) );
	//RT_CHECK_ERROR( rtContextRemoveVariable( context, temp_var ) );
	//RT_CHECK_ERROR( rtBufferDestroy( ambient_record_buffer ) );
	//RT_CHECK_ERROR( rtProgramDestroy( progs[1] ) );
	//RT_CHECK_ERROR( rtProgramDestroy( progs[0] ) );
	//RT_CHECK_ERROR( rtContextSetEntryPointCount( context, 1 ) );
}

void createAmbientSamplingCamera( RTcontext context, const VIEW view )
{
	RTprogram  ray_gen_program;
	RTprogram  exception_program;

	/* Ray generation program */
	sprintf( path_to_ptx, "%s/%s", sutilSamplesPtxDir(), "cuda_compile_ptx_generated_ambient_generator.cu.ptx" );
	RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "ambient_camera", &ray_gen_program ) );
	applyProgramVariable1ui( context, ray_gen_program, "camera", view.type ); // -vt
	applyProgramVariable3f( context, ray_gen_program, "eye", view.vp[0], view.vp[1], view.vp[2] ); // -vp
	applyProgramVariable3f( context, ray_gen_program, "U", view.hvec[0], view.hvec[1], view.hvec[2] );
	applyProgramVariable3f( context, ray_gen_program, "V", view.vvec[0], view.vvec[1], view.vvec[2] );
	applyProgramVariable3f( context, ray_gen_program, "W", view.vdir[0], view.vdir[1], view.vdir[2] ); // -vd
	applyProgramVariable2f( context, ray_gen_program, "fov", view.horiz, view.vert ); // -vh, -vv
	applyProgramVariable2f( context, ray_gen_program, "shift", view.hoff, view.voff ); // -vs, -vl
	applyProgramVariable2f( context, ray_gen_program, "clip", view.vfore, view.vaft ); // -vo, -va
	//applyProgramVariable1f( context, ray_gen_program, "dstrpix", dstrpix ); // -pj
	RT_CHECK_ERROR( rtContextSetRayGenerationProgram( context, AMBIENT_ENTRY, ray_gen_program ) );

	/* Exception program */
	//sprintf( path_to_ptx, "%s/%s", sutilSamplesPtxDir(), "cuda_compile_ptx_generated_ambient_generator.cu.ptx" );
	RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "exception", &exception_program ) );
	RT_CHECK_ERROR( rtContextSetExceptionProgram( context, AMBIENT_ENTRY, exception_program ) );

	/* Define ray types */
	applyContextVariable1ui( context, "ambient_record_ray_type", AMBIENT_RECORD_RAY );
}

void createGeometryInstanceAmbient( RTcontext context, RTgeometryinstance* instance )
{
	RTgeometry ambient_records;
	RTprogram  ambient_record_intersection_program;
	RTprogram  ambient_record_bounding_box_program;
	RTprogram  ambient_record_any_hit_program;
	RTprogram  ambient_record_miss_program;
	RTmaterial ambient_record_material;

	/* Create the geometry reference for OptiX. */
	RT_CHECK_ERROR( rtGeometryCreate( context, &ambient_records ) );
	RT_CHECK_ERROR( rtGeometrySetPrimitiveCount( ambient_records, nambvals ) );

	RT_CHECK_ERROR( rtMaterialCreate( context, &ambient_record_material ) );

	/* Define ray types */
	applyContextVariable1ui( context, "ambient_ray_type", AMBIENT_RAY );

	sprintf( path_to_ptx, "%s/%s", sutilSamplesPtxDir(), "cuda_compile_ptx_generated_ambient_records.cu.ptx" );
	RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "ambient_record_bounds", &ambient_record_intersection_program ) );
	RT_CHECK_ERROR( rtGeometrySetBoundingBoxProgram( ambient_records, ambient_record_intersection_program ) );
	RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "ambient_record_intersect", &ambient_record_bounding_box_program ) );
	RT_CHECK_ERROR( rtGeometrySetIntersectionProgram( ambient_records, ambient_record_bounding_box_program ) );
	RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "ambient_record_any_hit", &ambient_record_any_hit_program ) );
	RT_CHECK_ERROR( rtMaterialSetAnyHitProgram( ambient_record_material, AMBIENT_RAY, ambient_record_any_hit_program ) );

	/* Miss program */
	//sprintf( path_to_ptx, "%s/%s", sutilSamplesPtxDir(), "cuda_compile_ptx_generated_radiance_normal.cu.ptx" );
	RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "ambient_miss", &ambient_record_miss_program ) );
	RT_CHECK_ERROR( rtContextSetMissProgram( context, AMBIENT_RAY, ambient_record_miss_program ) );
	// TODO miss program could handle makeambient()

	/* Create the geometry instance containing the ambient records. */
	RT_CHECK_ERROR( rtGeometryInstanceCreate( context, instance ) );
	RT_CHECK_ERROR( rtGeometryInstanceSetGeometry( *instance, ambient_records ) );
	RT_CHECK_ERROR( rtGeometryInstanceSetMaterialCount( *instance, 1 ) );

	/* Apply this material to the geometry instance. */
	RT_CHECK_ERROR( rtGeometryInstanceSetMaterial( *instance, 0, ambient_record_material ) );
}

void populateAmbientRecords( RTcontext context )
{
	RTbuffer   ambient_record_buffer;
	void*      ambient_record_buffer_data;

	/* Create and map the buffers of ambient records. */
	createCustomBuffer1D( context, RT_BUFFER_INPUT, sizeof(AmbientRecord), nambvals, &ambient_record_buffer );
	RT_CHECK_ERROR( rtBufferMap( ambient_record_buffer, &ambient_record_buffer_data ) );

	/* Get the ambient records from the octree structure. */
	fprintf(stderr, "Ambient records %u\r\n", nambvals);
	gatherAmbientRecords(&atrunk, (AmbientRecord**)&ambient_record_buffer_data);

	/* Unmap the buffers of ambient records. */
	RT_CHECK_ERROR( rtBufferUnmap( ambient_record_buffer ) );
	applyContextObject( context, "ambient_records", ambient_record_buffer );
}

void gatherAmbientRecords(AMBTREE* at, AmbientRecord** records)
{
	AMBVAL* record;
	AMBTREE* child;

	unsigned int i;

	for (record = at->alist; record != NULL; record = record->next) {
		(*records)->pos.x = record->pos[0];
		(*records)->pos.y = record->pos[1];
		(*records)->pos.z = record->pos[2];

		(*records)->dir.x = record->dir[0];
		(*records)->dir.y = record->dir[1];
		(*records)->dir.z = record->dir[2];

		(*records)->gpos.x = record->gpos[0];
		(*records)->gpos.y = record->gpos[1];
		(*records)->gpos.z = record->gpos[2];

		(*records)->gdir.x = record->gdir[0];
		(*records)->gdir.y = record->gdir[1];
		(*records)->gdir.z = record->gdir[2];

		(*records)->val.x = record->val[0];
		(*records)->val.y = record->val[1];
		(*records)->val.z = record->val[2];

		(*records)->lvl = record->lvl;
		(*records)->weight = record->weight;
		(*records)->rad = record->rad;

		(*records)++;
	}

	child = at->kid;

	for (i = 0u; i < 8u; i++) {
		if (child != NULL) {
			gatherAmbientRecords(child++, records);
		}
	}
}

int saveAmbientRecords( AmbientRecord* record )
{
	AMBVAL amb;

	/* Check that an ambient record was created. */
	if ( record->rad < FTINY ) {
		if ( record->rad < -FTINY ) // Something bad happened
			fprintf(stderr, "Error on ambient: %f, %f, %f.\r\n", record->val.x, record->val.y, record->val.z );
		return(0);
	}

	amb.pos[0] = record->pos.x;
	amb.pos[1] = record->pos.y;
	amb.pos[2] = record->pos.z;

	amb.dir[0] = record->dir.x;
	amb.dir[1] = record->dir.y;
	amb.dir[2] = record->dir.z;

	amb.gpos[0] = record->gpos.x;
	amb.gpos[1] = record->gpos.y;
	amb.gpos[2] = record->gpos.z;

	amb.gdir[0] = record->gdir.x;
	amb.gdir[1] = record->gdir.y;
	amb.gdir[2] = record->gdir.z;

	amb.val[0] = record->val.x;
	amb.val[1] = record->val.y;
	amb.val[2] = record->val.z;

	amb.lvl = record->lvl;
	amb.weight = record->weight;
	amb.rad = record->rad;

	avsave(&amb);
	return(1);
}

RTacceleration acceleration;

void createAmbientAcceleration( RTcontext context, RTgeometryinstance instance )
{
	RTgeometrygroup geometrygroup;
	//RTacceleration  acceleration;

	/* Create a geometry group to hold the geometry instance.  This will be used as the top level group. */
	RT_CHECK_ERROR( rtGeometryGroupCreate( context, &geometrygroup ) );
	RT_CHECK_ERROR( rtGeometryGroupSetChildCount( geometrygroup, 1 ) );
	RT_CHECK_ERROR( rtGeometryGroupSetChild( geometrygroup, 0, instance ) );

	/* Set the geometry group as the top level object. */
	applyContextObject( context, "top_ambient", geometrygroup );

	/* create acceleration object for group and specify some build hints*/
	RT_CHECK_ERROR( rtAccelerationCreate( context, &acceleration ) );
	RT_CHECK_ERROR( rtAccelerationSetBuilder( acceleration, "Bvh" ) ); //TODO specify through command argument?
	RT_CHECK_ERROR( rtAccelerationSetTraverser( acceleration, "Bvh" ) );
	RT_CHECK_ERROR( rtGeometryGroupSetAcceleration( geometrygroup, acceleration ) );

	/* mark acceleration as dirty */
	RT_CHECK_ERROR( rtAccelerationMarkDirty( acceleration ) );
}

void iterativeAmbientCacheBuild( RTcontext context, RTgeometryinstance* instance, const VIEW view, const int width, const int height )
{
	//RTacceleration acceleration;
	RTgeometry     ambient_records;
	RTvariable     ambrec_read_var;
	RTvariable     ambounce_var;
	RTvariable     ambrec_var;
	RTvariable     level_var;
	RTbuffer       ambient_record_buffer;
	RTobject       ambrecs;
	AmbientRecord* ambient_record_buffer_data;
	
	unsigned int scaled_width, scaled_height, generated_record_count, useful_record_count, i, j;
	time_t kernel_start_time, kernel_end_time; // Timer in seconds for long jobs
	clock_t kernel_start_clock, kernel_end_clock; // Timer in clock cycles for short jobs

	scaled_width = width / optix_amb_scale;
	scaled_height = height / optix_amb_scale;
	generated_record_count = scaled_width * scaled_height * optix_amb_semgents;

	RT_CHECK_ERROR( rtContextQueryVariable( context, "ambounce", &ambounce_var ) );
	RT_CHECK_ERROR( rtVariableSet1i( ambounce_var, 0 ) );
	applyContextVariable1ui( context, "level", ambounce ); // Could be camera program variable

	createAmbientRecords( context, view, width, height );
	populateAmbientRecords( context );
	createGeometryInstanceAmbient( context, instance );
	createAmbientAcceleration( context, *instance );

	RT_CHECK_ERROR( rtGeometryInstanceGetGeometry( *instance, &ambient_records ) );

	RT_CHECK_ERROR( rtContextQueryVariable( context, "ambient_record_buffer", &ambrec_read_var ) );
	RT_CHECK_ERROR( rtVariableGetObject( ambrec_read_var, (RTobject*)&ambient_record_buffer ) );

	RT_CHECK_ERROR( rtContextQueryVariable( context, "level", &level_var ) );

	for ( i = 1u; i <= ambounce; i++ ) {
		// Create
		RT_CHECK_ERROR( rtVariableSet1i( ambounce_var, i ) );
		RT_CHECK_ERROR( rtVariableSet1ui( level_var, ambounce - i ) );



		ambdone();

		/* Run */
		RT_CHECK_ERROR( rtContextValidate( context ) );
		RT_CHECK_ERROR( rtContextCompile( context ) );
		kernel_start_time = time((time_t *)NULL);
		kernel_start_clock = clock();
		RT_CHECK_ERROR( rtContextLaunch2D( context, AMBIENT_ENTRY, scaled_width, scaled_height ) );
		kernel_end_clock = clock();
		kernel_end_time = time((time_t *)NULL);
		fprintf(stderr, "OptiX ambient kernel time: %u milliseconds.\r\n", (kernel_end_clock - kernel_start_clock) * 1000 / CLOCKS_PER_SEC);
		fprintf(stderr, "OptiX ambient kernel time: %u seconds.\r\n", kernel_end_time - kernel_start_time);

		RT_CHECK_ERROR( rtBufferMap( ambient_record_buffer, (void**)&ambient_record_buffer_data ) );
		RT_CHECK_ERROR( rtBufferUnmap( ambient_record_buffer ) );

		/* Copy the results to allocated memory. */
		//TODO the buffer could go directoy to creating Bvh
		useful_record_count = 0u;
		for (j = 0u; j < generated_record_count; j++) {
			useful_record_count += saveAmbientRecords( &ambient_record_buffer_data[j] );
//			ambient_record_buffer_data[j].rad = 0.0f; // Mark dirty
		}
		fprintf(stderr, "Retrieved %u ambient records from %u queries at level %u.\r\n", useful_record_count, generated_record_count, i);







		// Populate
		RT_CHECK_ERROR( rtContextQueryVariable( context, "ambient_records", &ambrec_var ) );
		RT_CHECK_ERROR( rtVariableGetObject( ambrec_var, &ambrecs ) );
		RT_CHECK_ERROR( rtBufferDestroy( (RTbuffer)ambrecs ) );
		RT_CHECK_ERROR( rtContextRemoveVariable( context, ambrec_var ) );
		populateAmbientRecords( context );

		// Geom
		RT_CHECK_ERROR( rtGeometrySetPrimitiveCount( ambient_records, nambvals ) );

		// Accel
		RT_CHECK_ERROR( rtAccelerationMarkDirty( acceleration ) );
	}

}