/*
 * Copyright (c) 2013 Nathaniel Jones
 * Massachusetts Institute of Technology
 */

//#ifdef OPTIX
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <standard.h> /* TODO just get the includes that are required? */
#include <object.h>
#include <otypes.h>
#include "ray.h"
#include "source.h"
#include "ambient.h"
#include <face.h>

#include "optix_radiance.h"
#include "commonStructs.h"

#define RADIANCE_ENTRY 0u

#define RADIANCE_RAY 0u
#define SHADOW_RAY   1u
#define AMBIENT_RECORD_RAY 3u


void printDeviceInfo();
void createContext( RTcontext* context, RTbuffer* buffer, const int width, const int height, const double alarm );
void applyRadianceSettings( RTcontext context );
void createCamera( RTcontext context, const VIEW view );
void createNormalMaterial( RTcontext context, RTgeometryinstance instance, int index, OBJREC* rec );
void createGlassMaterial( RTcontext context, RTgeometryinstance instance, int index, OBJREC* rec );
void createLightMaterial( RTcontext context, RTgeometryinstance instance, int index, OBJREC* rec );
void createGeometryInstance( RTcontext context, RTgeometryinstance* instance );
void createAcceleration( RTcontext context, RTgeometryinstance instance );
void countObjectsInScene(unsigned int* vertex_count, unsigned int* triangle_count, unsigned int* material_count, unsigned int* point_light_count, unsigned int* sky_bright_count, unsigned int* perez_lum_count);
int getSkyType( OBJREC* bright_func );


extern COLOR* color_output = NULL;
extern float* depth_output = NULL;

/* from rpict.c */
extern double  dstrpix;			/* square pixel distribution */
/* from ambient.c */
extern double  avsum;		/* computed ambient value sum (log) */
extern unsigned int  navsum;	/* number of values in avsum */
extern unsigned int  nambvals;	/* total number of indirect values */

static int use_ambient = 0;
static int calc_ambient = 0;


extern void renderOptix(const VIEW view, const int width, const int height, const double alarm, COLOR* colors, float* depths)
{
	/* Primary RTAPI objects */
	RTcontext           context;
	RTbuffer            output_buffer;
	RTgeometryinstance  instance;
	RTgeometryinstance  ambient_records;

	/* Parameters */
	unsigned int i, size, milliseconds;
	float* data;
	time_t kernel_start_time, kernel_end_time; // Timer in seconds for long jobs
	clock_t kernel_start_clock, kernel_end_clock; // Timer in clock cycles for short jobs

	/* Set the size */
	size = width * height;

	/* Check if irradiance cache is used */
	use_ambient = ambacc > FTINY;
	calc_ambient = use_ambient && nambvals == 0u;// && (ambfile == NULL || !ambfile[0]); // TODO Should really look at ambfp in ambinet.c to check that file is readable

	/* Setup state */
	printDeviceInfo();
	createContext( &context, &output_buffer, width, height, alarm );
	applyRadianceSettings( context );
	createCamera( context, view );
	createGeometryInstance( context, &instance );
	createAcceleration( context, instance );

	/* Set up irradiance cache of ambient values */
	if ( use_ambient ) { // Don't bother with ambient records if -aa is set to zero
		if ( calc_ambient ) // Run pre-process only if no ambient file is available
			createAmbientRecords( context, view, width, height );
		populateAmbientRecords( context );
		createGeometryInstanceAmbient( context, &ambient_records );
		createAmbientAcceleration( context, ambient_records );
		//iterativeAmbientCacheBuild( context, &ambient_records, view, width, height );
	}

	/* Run */
	RT_CHECK_ERROR( rtContextValidate( context ) );
	RT_CHECK_ERROR( rtContextCompile( context ) );
	kernel_start_time = time((time_t *)NULL);
	kernel_start_clock = clock();
	RT_CHECK_ERROR( rtContextLaunch2D( context, RADIANCE_ENTRY, width, height ) );
	kernel_end_clock = clock();
	kernel_end_time = time((time_t *)NULL);
	fprintf(stderr, "OptiX radiance kernel time: %u milliseconds.\r\n", (kernel_end_clock - kernel_start_clock) * 1000 / CLOCKS_PER_SEC);
	fprintf(stderr, "OptiX radiance kernel time: %u seconds.\r\n", kernel_end_time - kernel_start_time);

	RT_CHECK_ERROR( rtBufferMap( output_buffer, (void**)&data ) );
	RT_CHECK_ERROR( rtBufferUnmap( output_buffer ) );

	/* Copy the results to allocated memory. */
	for (i = 0u; i < size; i++) {
		copycolor( colors[i], data );
		depths[i] = data[3];
		data += 4;
	}

	/* Clean up */
	RT_CHECK_ERROR( rtContextDestroy( context ) );
}

void printDeviceInfo()
{
	unsigned int version, device_count;
	unsigned int i;
	unsigned int multiprocessor_count, threads_per_block, clock_rate, texture_count, timeout_enabled, tcc_driver, cuda_device;
	unsigned int compute_capability[2];
	char device_name[128];
	RTsize memory_size;

	rtGetVersion( &version );
	rtDeviceGetDeviceCount( &device_count );
	fprintf(stderr, "Starting OptiX %i on %i devices:\r\n", version, device_count);

	for (i = 0; i < device_count; i++) {
		rtDeviceGetAttribute( i, RT_DEVICE_ATTRIBUTE_NAME, sizeof(device_name), &device_name );
		rtDeviceGetAttribute( i, RT_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT, sizeof(multiprocessor_count), &multiprocessor_count );
		rtDeviceGetAttribute( i, RT_DEVICE_ATTRIBUTE_MAX_THREADS_PER_BLOCK, sizeof(threads_per_block), &threads_per_block );
		rtDeviceGetAttribute( i, RT_DEVICE_ATTRIBUTE_CLOCK_RATE, sizeof(clock_rate), &clock_rate );
		rtDeviceGetAttribute( i, RT_DEVICE_ATTRIBUTE_TOTAL_MEMORY, sizeof(memory_size), &memory_size );
		rtDeviceGetAttribute( i, RT_DEVICE_ATTRIBUTE_MAX_HARDWARE_TEXTURE_COUNT, sizeof(texture_count), &texture_count );
		rtDeviceGetAttribute( i, RT_DEVICE_ATTRIBUTE_COMPUTE_CAPABILITY, sizeof(compute_capability), &compute_capability );
		rtDeviceGetAttribute( i, RT_DEVICE_ATTRIBUTE_EXECUTION_TIMEOUT_ENABLED, sizeof(timeout_enabled), &timeout_enabled );
		rtDeviceGetAttribute( i, RT_DEVICE_ATTRIBUTE_TCC_DRIVER, sizeof(tcc_driver), &tcc_driver );
		rtDeviceGetAttribute( i, RT_DEVICE_ATTRIBUTE_CUDA_DEVICE_ORDINAL, sizeof(cuda_device), &cuda_device );
		fprintf(stderr, "Device %u: %s with %u multiprocessors, %u threads per block, %u Hz, %u Bytes memory, %u hardware textures, compute capability %u.%u, %u timeout, Tesla compute cluster driver %u, cuda device %u.\r\n",
			i, device_name, multiprocessor_count, threads_per_block, clock_rate, memory_size, texture_count, compute_capability[0], compute_capability[1], timeout_enabled, tcc_driver, cuda_device);
	}
	fprintf(stderr, "\r\n");
}

void createContext( RTcontext* context, RTbuffer* output_buffer, const int width, const int height, const double alarm )
{
	//RTbuffer seed_buffer;

	//unsigned int* seeds;
	//unsigned int i;

	/* Setup context */
	RT_CHECK_ERROR2( rtContextCreate( context ) );
	RT_CHECK_ERROR2( rtContextSetRayTypeCount( *context, 2 + use_ambient + calc_ambient ) ); /* shadow and radiance, potentially ambient */
	RT_CHECK_ERROR2( rtContextSetEntryPointCount( *context, 1 + calc_ambient ) );

	/* Set stack size for GPU threads */
	if (optix_stack_size > 0)
		RT_CHECK_ERROR2( rtContextSetStackSize( *context, optix_stack_size ) );

	/* Create a buffer of random number seeds */
	//createBuffer2D( *context, RT_BUFFER_INPUT, RT_FORMAT_UNSIGNED_INT, width, height, &seed_buffer );
	//RT_CHECK_ERROR2( rtBufferMap( seed_buffer, (void**)&seeds ) );
	//for ( i = 0; i < width * height; ++i ) 
	//	seeds[i] = rand();
	//RT_CHECK_ERROR2( rtBufferUnmap( seed_buffer ) );
	//applyContextObject( *context, "rnd_seeds", seed_buffer );

	/* Render result buffer */
	createBuffer2D( *context, RT_BUFFER_OUTPUT, RT_FORMAT_FLOAT4, width, height, output_buffer );
	applyContextObject( *context, "output_buffer", *output_buffer );

#ifdef TIMEOUT_CALLBACK
	if (alarm > 0)
		RT_CHECK_ERROR2( rtContextSetTimeoutCallback( *context, timeoutCallback, alarm ) );
#endif

#ifdef DEBUG_OPTIX
	/* Enable exception checking */
	RT_CHECK_ERROR2( rtContextSetExceptionEnabled( *context, RT_EXCEPTION_ALL, 1 ) );
#endif
}

void applyRadianceSettings( RTcontext context )
{
	/* Define ray types */
	applyContextVariable1ui( context, "radiance_ray_type", RADIANCE_RAY );
	applyContextVariable1ui( context, "shadow_ray_type", SHADOW_RAY );

	/* Set hard coded parameters */
	applyContextVariable3f( context, "CIE_rgbf", CIE_rf, CIE_gf, CIE_bf ); // from color.h

	/* Set direct parameters */
	//applyContextVariable1f( context, "dstrsrc", dstrsrc ); // -dj
	//applyContextVariable1f( context, "srcsizerat", srcsizerat ); // -ds
	//applyContextVariable1f( context, "shadthresh", shadthresh ); // -dt
	//applyContextVariable1f( context, "shadcert", shadcert ); // -dc
	//applyContextVariable1i( context, "directrelay", directrelay ); // -dr
	//applyContextVariable1i( context, "vspretest", vspretest ); // -dp
	//applyContextVariable1i( context, "directvis", directvis ); // -dv

	/* Set specular parameters */
	applyContextVariable1f( context, "specthresh", specthresh ); // -st
	applyContextVariable1f( context, "specjitter", specjitter ); // -ss

	/* Set ambient parameters */
	applyContextVariable3f( context, "ambval", ambval[0], ambval[1], ambval[2] ); // -av
	applyContextVariable1i( context, "ambvwt", ambvwt ); // -aw, zero by default
	applyContextVariable1i( context, "ambounce", ambounce ); // -ab
	//applyContextVariable1i( context, "ambres", ambres ); // -ar
	applyContextVariable1f( context, "ambacc", ambacc ); // -aa
	applyContextVariable1i( context, "ambdiv", ambdiv ); // -ad
	applyContextVariable1i( context, "ambssamp", ambssamp ); // -as
	applyContextVariable1f( context, "maxarad", maxarad ); // maximum ambient radius from ambient.h, based on armbres
	applyContextVariable1f( context, "minarad", minarad ); // minimum ambient radius from ambient.h, based on armbres
	applyContextVariable1f( context, "avsum", avsum ); // computed ambient value sum (log) from ambient.c
	applyContextVariable1ui( context, "navsum", navsum ); // number of values in avsum from ambient.c

	/* Set medium parameters */
	//applyContextVariable3f( context, "cextinction", cextinction[0], cextinction[1], cextinction[2] ); // -me
	//applyContextVariable3f( context, "salbedo", salbedo[0], salbedo[1], salbedo[2] ); // -me
	//applyContextVariable1f( context, "seccg", seccg ); // -mg
	//applyContextVariable1f( context, "ssampdist", ssampdist ); // -ms

	/* Set ray limitting parameters */
	applyContextVariable1f( context, "minweight", minweight ); // -lw, from ray.h
	applyContextVariable1i( context, "maxdepth", maxdepth ); // -lr, from ray.h, negative values indicate Russian roulette
}

void createCamera( RTcontext context, const VIEW view )
{
	RTprogram  ray_gen_program;
	RTprogram  exception_program;
	RTprogram  miss_program;
	RTprogram  miss_shadow_program;

	/* Ray generation program */
	sprintf( path_to_ptx, "%s/%s", sutilSamplesPtxDir(), "cuda_compile_ptx_generated_pinhole_camera.cu.ptx" );
	RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "image_camera", &ray_gen_program ) );
	applyProgramVariable1ui( context, ray_gen_program, "camera", view.type ); // -vt
	applyProgramVariable3f( context, ray_gen_program, "eye", view.vp[0], view.vp[1], view.vp[2] ); // -vp
	applyProgramVariable3f( context, ray_gen_program, "U", view.hvec[0], view.hvec[1], view.hvec[2] );
	applyProgramVariable3f( context, ray_gen_program, "V", view.vvec[0], view.vvec[1], view.vvec[2] );
	applyProgramVariable3f( context, ray_gen_program, "W", view.vdir[0], view.vdir[1], view.vdir[2] ); // -vd
	applyProgramVariable2f( context, ray_gen_program, "fov", view.horiz, view.vert ); // -vh, -vv
	applyProgramVariable2f( context, ray_gen_program, "shift", view.hoff, view.voff ); // -vs, -vl
	applyProgramVariable2f( context, ray_gen_program, "clip", view.vfore, view.vaft ); // -vo, -va
	applyProgramVariable1f( context, ray_gen_program, "dstrpix", dstrpix ); // -pj
	RT_CHECK_ERROR( rtContextSetRayGenerationProgram( context, RADIANCE_ENTRY, ray_gen_program ) );

	/* Exception program */
	//sprintf( path_to_ptx, "%s/%s", sutilSamplesPtxDir(), "cuda_compile_ptx_generated_pinhole_camera.cu.ptx" );
	RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "exception", &exception_program ) );
	RT_CHECK_ERROR( rtContextSetExceptionProgram( context, RADIANCE_ENTRY, exception_program ) );

	/* Miss program */
	sprintf( path_to_ptx, "%s/%s", sutilSamplesPtxDir(), "cuda_compile_ptx_generated_sky_background.cu.ptx" );
	RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "miss", &miss_program ) );
	RT_CHECK_ERROR( rtContextSetMissProgram( context, RADIANCE_RAY, miss_program ) );
	RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "miss_shadow", &miss_shadow_program ) );
	RT_CHECK_ERROR( rtContextSetMissProgram( context, SHADOW_RAY, miss_shadow_program ) );
}

/* Count the number of vertices and triangles belonging to all faces in the scene. */
void countObjectsInScene( unsigned int* vertex_count, unsigned int* triangle_count, unsigned int* material_count, unsigned int* distant_light_count, unsigned int* sky_bright_count, unsigned int* perez_lum_count )
{
	OBJREC* rec;
	FACE face;
	unsigned int i, v_count, t_count, m_count, dl_count, cie_count, perez_count;
	v_count = t_count = m_count = dl_count = cie_count = perez_count = 0;

	for (i = 0; i < nobjects; i++) {
		rec = objptr(i);
		//if (issurface(rec.otype)) {
		switch (rec->otype) {
		case MAT_PLASTIC: // Plastic material
		case MAT_METAL: // Metal material
		case MAT_TRANS: // Translucent material
		case MAT_GLASS: // Glass material
		case MAT_DIELECTRIC: // Dielectric material
		case MAT_LIGHT: // primary light source material, may modify a face or a source (solid angle)
			m_count++;
			break;
		case OBJ_FACE: // Typical polygons
			face = *getface(rec);
			v_count += face.nv;
			t_count += face.nv - 2;
			break;
		case OBJ_SOURCE: // The sun, for example
			dl_count++;
			break;
		case PAT_BFUNC: // brightness function, used for sky brightness
			// TODO come up with a more general way for handling functions
			switch (getSkyType(rec)) {
			case SKY_CIE: // CIE sky
				cie_count++;
				break;
			case SKY_PEREZ: // Perez sky
				perez_count++;
				break;
			default:
				break;
			}
			break;
		default:
			break;
		}
	}

	*vertex_count = v_count;
	*triangle_count = t_count;
	*material_count = m_count;
	*distant_light_count = dl_count;
	*sky_bright_count = cie_count;
	*perez_lum_count = perez_count;
}

int getSkyType( OBJREC* rec )
{
	if ( rec->oargs.nsargs == 2 ) {
		if ( !strcmp(rec->oargs.sarg[0], "skybr") && !strcmp(rec->oargs.sarg[1], "skybright.cal") )
			return SKY_CIE;
		if ( !strcmp(rec->oargs.sarg[0], "skybright") && !strcmp(rec->oargs.sarg[1], "perezlum.cal") )
			return SKY_PEREZ;
	}
	return SKY_NONE;
}

void createGeometryInstance( RTcontext context, RTgeometryinstance* instance )
{
	RTgeometry mesh;
	RTprogram  mesh_intersection_program;
	RTprogram  mesh_bounding_box_program;

	RTbuffer   light_buffer;
	void*      light_buffer_data;

	RTbuffer   cie_buffer;
	void*      cie_buffer_data;

	RTbuffer   perez_buffer;
	void*      perez_buffer_data;

	RTbuffer   vertex_buffer;
	RTbuffer   normal_buffer;
	RTbuffer   tex_coord_buffer;
	RTbuffer   vertex_index_buffer;
	RTbuffer   normal_index_buffer;
	RTbuffer   tex_coord_index_buffer;
	RTbuffer   material_index_buffer;
	float*     vertex_buffer_data;
	float*     normal_buffer_data;
	float*     tex_coord_buffer_data;
	int*       v_index_buffer_data;
	int*       n_index_buffer_data;
	int*       t_index_buffer_data;
	unsigned int* m_index_buffer_data;

	int i, j, t;
	unsigned int vi, ni, ti, mi, vii, nii, tii, mii, dli, sbi, pli;
	unsigned int vertex_count, triangle_count, material_count, distant_light_count, sky_bright_count, perez_lum_count;
	OBJREC* rec, *parent;
	FACE* face;
	SRCREC source;
	BasicLight light;
	SkyBright cie;
	PerezLum perez;

	/* This array gives the OptiX buffer index of each rad file object.
	 * The OptiX buffer refered to depends on the type of object.
	 * If the object does not have an index in an OptiX buffer, -1 is given. */
	int* buffer_entry_index;

	buffer_entry_index = (int *)malloc(sizeof(int) * nobjects);

	/* Count the number of each type of object in the scene. */
	countObjectsInScene(&vertex_count, &triangle_count, &material_count, &distant_light_count, &sky_bright_count, &perez_lum_count);

	/* Create the geometry reference for OptiX. */
	RT_CHECK_ERROR( rtGeometryCreate( context, &mesh ) );
	RT_CHECK_ERROR( rtGeometrySetPrimitiveCount( mesh, triangle_count ) );

	sprintf( path_to_ptx, "%s/%s", sutilSamplesPtxDir(), "cuda_compile_ptx_generated_triangle_mesh.cu.ptx" );
	RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "mesh_bounds", &mesh_bounding_box_program ) );
	RT_CHECK_ERROR( rtGeometrySetBoundingBoxProgram( mesh, mesh_bounding_box_program ) );
	RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "mesh_intersect", &mesh_intersection_program ) );
	RT_CHECK_ERROR( rtGeometrySetIntersectionProgram( mesh, mesh_intersection_program ) );
	applyProgramVariable1ui( context, mesh_intersection_program, "backvis", backvis ); // -bv

	/* Create the geometry instance containing the geometry. */
	RT_CHECK_ERROR( rtGeometryInstanceCreate( context, instance ) );
	RT_CHECK_ERROR( rtGeometryInstanceSetGeometry( *instance, mesh ) );
	RT_CHECK_ERROR( rtGeometryInstanceSetMaterialCount( *instance, material_count ) );

	/* Create buffers for storing lighting information. */
	createCustomBuffer1D( context, RT_BUFFER_INPUT, sizeof(BasicLight), distant_light_count, &light_buffer );
	RT_CHECK_ERROR( rtBufferMap( light_buffer, &light_buffer_data ) );

	createCustomBuffer1D( context, RT_BUFFER_INPUT, sizeof(SkyBright), sky_bright_count, &cie_buffer );
	RT_CHECK_ERROR( rtBufferMap( cie_buffer, &cie_buffer_data ) );

	createCustomBuffer1D( context, RT_BUFFER_INPUT, sizeof(PerezLum), perez_lum_count, &perez_buffer );
	RT_CHECK_ERROR( rtBufferMap( perez_buffer, &perez_buffer_data ) );

	/* Create buffers for storing geometry information. */
	createBuffer1D( context, RT_BUFFER_INPUT, RT_FORMAT_FLOAT3, vertex_count, &vertex_buffer );
	RT_CHECK_ERROR( rtBufferMap( vertex_buffer, (void**)&vertex_buffer_data ) );

	createBuffer1D( context, RT_BUFFER_INPUT, RT_FORMAT_FLOAT3, vertex_count, &normal_buffer );
	RT_CHECK_ERROR( rtBufferMap( normal_buffer, (void**)&normal_buffer_data ) );

	createBuffer1D( context, RT_BUFFER_INPUT, RT_FORMAT_FLOAT2, vertex_count, &tex_coord_buffer );
	RT_CHECK_ERROR( rtBufferMap( tex_coord_buffer, (void**)&tex_coord_buffer_data ) );

	createBuffer1D( context, RT_BUFFER_INPUT, RT_FORMAT_INT3, triangle_count, &vertex_index_buffer );
	RT_CHECK_ERROR( rtBufferMap( vertex_index_buffer, (void**)&v_index_buffer_data ) );

	createBuffer1D( context, RT_BUFFER_INPUT, RT_FORMAT_INT3, triangle_count, &normal_index_buffer );
	RT_CHECK_ERROR( rtBufferMap( normal_index_buffer, (void**)&n_index_buffer_data ) );

	createBuffer1D( context, RT_BUFFER_INPUT, RT_FORMAT_INT3, triangle_count, &tex_coord_index_buffer );
	RT_CHECK_ERROR( rtBufferMap( tex_coord_index_buffer, (void**)&t_index_buffer_data ) );

	createBuffer1D( context, RT_BUFFER_INPUT, RT_FORMAT_UNSIGNED_INT, triangle_count, &material_index_buffer );
	RT_CHECK_ERROR( rtBufferMap( material_index_buffer, (void**)&m_index_buffer_data ) );

	vi = ni = ti = mi = vii = nii = tii = mii = dli = sbi = pli = 0u;

	/* Get the scene geometry as a list of triangles. */
	fprintf(stderr, "Num objects %i\r\n", nobjects);
	for (i = 0; i < nobjects; i++) {
		/* By default, no buffer entry is refered to. */
		buffer_entry_index[i] = -1;

		rec = objptr(i);
		//fprintf(stderr, "Object %i: %s\r\n", i, rec.oname);
		//fprintf(stderr, "Object %i modifies: %i\r\n", i, rec.omod);
		//fprintf(stderr, "Object %i type: %i\r\n", i, rec.otype);
		//fprintf(stderr, "Object %i structure: %s\r\n", i, rec.os);

		//if (issurface(rec.otype)) {
		switch(rec->otype) {
		case MAT_PLASTIC: // Plastic material
		case MAT_METAL: // Metal material
		case MAT_TRANS: // Translucent material
			buffer_entry_index[i] = mi;
			createNormalMaterial( context, *instance, mi++, rec );
			break;
		case MAT_GLASS: // Glass material
		case MAT_DIELECTRIC: // Dielectric material TODO handle separately, see dialectric.c
			buffer_entry_index[i] = mi;
			createGlassMaterial( context, *instance, mi++, rec );
			break;
		case MAT_LIGHT: // primary light source material, may modify a face or a source (solid angle)
			buffer_entry_index[i] = mi;
			createLightMaterial( context, *instance, mi++, rec );
			break;
		case OBJ_FACE:
			face = getface(rec);

			/* Write the indices to the buffers */
			for (j = 2; j < face->nv; j++) {
				t = vi / 3;
				v_index_buffer_data[vii++] = t;
				v_index_buffer_data[vii++] = t + j - 1;
				v_index_buffer_data[vii++] = t + j;

				n_index_buffer_data[nii++] = t;
				n_index_buffer_data[nii++] = t + j - 1;
				n_index_buffer_data[nii++] = t + j;

				t_index_buffer_data[tii++] = t;
				t_index_buffer_data[tii++] = t + j - 1;
				t_index_buffer_data[tii++] = t + j;

				//m_index_buffer_data[mii++] = buffer_entry_index[rec.omod]; //This may be the case once texture functions are implemented
				m_index_buffer_data[mii++] = buffer_entry_index[objndx(findmaterial(rec))]; // internal calls seem redundant
			}

			/* Write the vertices to the buffers */
			for (j = 0; j < face->nv; j++) {
				vertex_buffer_data[vi++] = face->va[3 * j]; //TODO this only works for convex polygons
				vertex_buffer_data[vi++] = face->va[3 * j + 1];
				vertex_buffer_data[vi++] = face->va[3 * j + 2];

				//TODO Implement bump maps from texfunc and texdata
				// This might be done in the Intersecion program in triangle_mesh.cu rather than here
				normal_buffer_data[ni++] = face->norm[0];
				normal_buffer_data[ni++] = face->norm[1];
				normal_buffer_data[ni++] = face->norm[2];

				//TODO Implement texture maps from colorfunc, brightfunc, colordata, brightdata, and colorpict
				// This might be done in the Intersecion program in triangle_mesh.cu rather than here
				tex_coord_buffer_data[ti++] = 0.0f;
				tex_coord_buffer_data[ti++] = 0.0f;
			}
			break;
		case OBJ_SOURCE:
			ssetsrc(&source, rec);
			parent = findmaterial(rec);
			light.color.x = parent->oargs.farg[0]; // TODO these are given in RGB radiance value (watts/steradian/m2)
			light.color.y = parent->oargs.farg[1];
			light.color.z = parent->oargs.farg[2];
			light.pos.x   = source.sloc[0];
			light.pos.y   = source.sloc[1];
			light.pos.z   = source.sloc[2];
			light.solid_angle = source.ss2;
			if (parent->oargs.nfargs > 3 && parent->oargs.farg[3] == 0) {
				light.casts_shadow = 0;
			} else {
				light.casts_shadow = 1;
			}

			/* Check for a parent function. */
			if (parent->omod > -1) {
				light.type = getSkyType(objptr(parent->omod));
				light.function = buffer_entry_index[parent->omod];
			}

			//fprintf(stderr, "Source [%f, %f, %f]\r\n", source.sloc[0], source.sloc[1], source.sloc[2]);
			//fprintf(stderr, "Color [%f, %f, %f]\r\n", parent.oargs.farg[0], parent.oargs.farg[1], parent.oargs.farg[2]);
			//fprintf(stderr, "Solid Angle %f\r\n", source.ss2);
			//fprintf(stderr, "DimU [%f, %f, %f]\r\n", source.ss[0][0], source.ss[0][1], source.ss[0][2]);
			//fprintf(stderr, "DimV [%f, %f, %f]\r\n", source.ss[1][0], source.ss[1][1], source.ss[1][2]);
			//fprintf(stderr, "DimW [%f, %f, %f]\r\n", source.ss[2][0], source.ss[2][1], source.ss[2][2]);
			((BasicLight*)light_buffer_data)[dli++] = light;
			break;
		case PAT_BFUNC: // brightness function, used for sky brightness
			switch (getSkyType(rec)) {
			case SKY_CIE: // CIE sky
				buffer_entry_index[i] = sbi;

				cie.type = rec->oargs.farg[0];
				cie.zenith = rec->oargs.farg[1];
				cie.ground = rec->oargs.farg[2];
				cie.factor = rec->oargs.farg[3];
				cie.sun.x = rec->oargs.farg[4];
				cie.sun.y = rec->oargs.farg[5];
				cie.sun.z = rec->oargs.farg[6];
				((SkyBright*)cie_buffer_data)[sbi++] = cie;
				break;
			case SKY_PEREZ: // Perez sky
				buffer_entry_index[i] = pli;

				perez.diffuse = rec->oargs.farg[0];
				perez.ground = rec->oargs.farg[1];
				for (j = 0; j < 5; j++) {
					perez.coef[j] = rec->oargs.farg[j + 2];
				}
				perez.sun.x = rec->oargs.farg[7];
				perez.sun.y = rec->oargs.farg[8];
				perez.sun.z = rec->oargs.farg[9];
				((PerezLum*)perez_buffer_data)[pli++] = perez;
				break;
			default:
				break;
			}
			break;
		case TEX_FUNC:
			break;
		default:
			fprintf(stderr, "Object %i: %s\r\n", i, rec->oname);
			fprintf(stderr, "Object %i modifies: %i\r\n", i, rec->omod);
			fprintf(stderr, "Object %i type: %i\r\n", i, rec->otype);
			fprintf(stderr, "Object %i structure: %s\r\n", i, rec->os);

			for (j = 0; j < rec->oargs.nsargs; j++) {
				fprintf(stderr, "S Arg %i: %s\r\n", j, rec->oargs.sarg[j]);
			}
			for (j = 0; j < rec->oargs.nfargs; j++) {
				fprintf(stderr, "R Arg %i: %f\r\n", j, rec->oargs.farg[j]);
			}

			fprintf(stderr, "\r\n");
			break;
		}
		//fprintf(stderr, "\r\n");
	}

	free( buffer_entry_index );

	/* Unmap and apply the lighting buffer. */
	RT_CHECK_ERROR( rtBufferUnmap( light_buffer ) );
	applyContextObject( context, "lights", light_buffer );

	RT_CHECK_ERROR( rtBufferUnmap( cie_buffer ) );
	applyContextObject( context, "sky_brights", cie_buffer );

	RT_CHECK_ERROR( rtBufferUnmap( perez_buffer ) );
	applyContextObject( context, "perez_lums", perez_buffer );

	/* Unmap and apply the geometry buffers. */
	RT_CHECK_ERROR( rtBufferUnmap( vertex_buffer ) );
	applyGeometryObject( context, mesh, "vertex_buffer", vertex_buffer );

	RT_CHECK_ERROR( rtBufferUnmap( normal_buffer ) );
	applyGeometryObject( context, mesh, "normal_buffer", normal_buffer );

	RT_CHECK_ERROR( rtBufferUnmap( tex_coord_buffer ) );
	applyGeometryObject( context, mesh, "texcoord_buffer", tex_coord_buffer );

	RT_CHECK_ERROR( rtBufferUnmap( vertex_index_buffer ) );
	applyGeometryObject( context, mesh, "vindex_buffer", vertex_index_buffer );

	RT_CHECK_ERROR( rtBufferUnmap( normal_index_buffer ) );
	applyGeometryObject( context, mesh, "nindex_buffer", normal_index_buffer );

	RT_CHECK_ERROR( rtBufferUnmap( tex_coord_index_buffer ) );
	applyGeometryObject( context, mesh, "tindex_buffer", tex_coord_index_buffer );

	RT_CHECK_ERROR( rtBufferUnmap( material_index_buffer ) );
	applyGeometryObject( context, mesh, "material_buffer", material_index_buffer );
}

void createNormalMaterial( RTcontext context, RTgeometryinstance instance, int index, OBJREC* rec )
{
	RTmaterial material;
	RTprogram closest_hit_program;
	RTprogram any_hit_program;
	RTprogram ambient_hit_program;

	unsigned int type;

	/* Create our hit programs to be shared among all materials */
	sprintf( path_to_ptx, "%s/%s", sutilSamplesPtxDir(), "cuda_compile_ptx_generated_radiance_normal.cu.ptx" );
	RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "closest_hit_radiance", &closest_hit_program ) );
	RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "any_hit_shadow", &any_hit_program ) );

	RT_CHECK_ERROR( rtMaterialCreate( context, &material ) );

	/* Note that we are leaving anyHitProgram[0] and closestHitProgram[1] as NULL.
	 * This is because our radiance rays only need closest_hit and shadow rays only
	 * need any_hit */
	RT_CHECK_ERROR( rtMaterialSetClosestHitProgram( material, RADIANCE_RAY, closest_hit_program ) );
	RT_CHECK_ERROR( rtMaterialSetAnyHitProgram( material, SHADOW_RAY, any_hit_program ) );

	if ( calc_ambient ) {
		sprintf( path_to_ptx, "%s/%s", sutilSamplesPtxDir(), "cuda_compile_ptx_generated_ambient_normal.cu.ptx" );
		RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "closest_hit_ambient", &ambient_hit_program ) );
		RT_CHECK_ERROR( rtMaterialSetClosestHitProgram( material, AMBIENT_RECORD_RAY, ambient_hit_program ) );
	}

	/* Check type of material */
	switch(rec->otype) {
		case MAT_PLASTIC: type = SHADE_MAT_PLASTIC; break; // Plastic material
		case MAT_METAL: type = SHADE_MAT_METAL; break; // Metal material
		case MAT_TRANS: type = SHADE_MAT_TRANS; break; // Translucent material
	}

	/* Set variables to be consumed by material for this geometry instance */
	applyMaterialVariable1ui( context, material, "type", type );
	applyMaterialVariable3f( context, material, "color", rec->oargs.farg[0], rec->oargs.farg[1], rec->oargs.farg[2] );
	applyMaterialVariable1f( context, material, "spec", rec->oargs.farg[3] );
	applyMaterialVariable1f( context, material, "rough", rec->oargs.farg[4] );
	if (type == SHADE_MAT_TRANS) { // it's a translucent material
		applyMaterialVariable1f( context, material, "trans", rec->oargs.farg[5] );
		applyMaterialVariable1f( context, material, "tspec", rec->oargs.farg[6] );
	}

	/* Apply this material to the geometry instance. */
	RT_CHECK_ERROR( rtGeometryInstanceSetMaterial( instance, index, material ) );
}

void createGlassMaterial( RTcontext context, RTgeometryinstance instance, int index, OBJREC* rec )
{
	RTmaterial material;
	RTprogram closest_hit_program;
	RTprogram closest_hit_shadow_program;
	RTprogram ambient_hit_program;

	//float rindex;

	/* Create our hit programs to be shared among all materials */
	sprintf( path_to_ptx, "%s/%s", sutilSamplesPtxDir(), "cuda_compile_ptx_generated_radiance_glass.cu.ptx" );
	RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "closest_hit_radiance", &closest_hit_program ) );
	RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "closest_hit_shadow", &closest_hit_shadow_program ) );

	RT_CHECK_ERROR( rtMaterialCreate( context, &material ) );

	/* Note that we are leaving anyHitProgram[0] and closestHitProgram[1] as NULL.
	 * This is because our radiance rays only need closest_hit and shadow rays only
	 * need any_hit */
	RT_CHECK_ERROR( rtMaterialSetClosestHitProgram( material, RADIANCE_RAY, closest_hit_program ) );
	RT_CHECK_ERROR( rtMaterialSetClosestHitProgram( material, SHADOW_RAY, closest_hit_shadow_program ) );

	if ( calc_ambient ) {
		sprintf( path_to_ptx, "%s/%s", sutilSamplesPtxDir(), "cuda_compile_ptx_generated_ambient_normal.cu.ptx" );
		RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "any_hit_ambient_glass", &ambient_hit_program ) );
		RT_CHECK_ERROR( rtMaterialSetAnyHitProgram( material, AMBIENT_RECORD_RAY, ambient_hit_program ) );
	}

	/* Set variables to be consumed by material for this geometry instance */
	applyMaterialVariable3f( context, material, "color", rec->oargs.farg[0], rec->oargs.farg[1], rec->oargs.farg[2] );
	if (rec->oargs.nfargs > 3) {
		applyMaterialVariable1f( context, material, "rindex", rec->oargs.farg[3] );
		//rindex = rec.oargs.farg[3];
	//} else {
	//	rindex = 1.52f;
	}
	//applyMaterialVariable1f( context, material, "rindex", rindex );

	/* Apply this material to the geometry instance. */
	RT_CHECK_ERROR( rtGeometryInstanceSetMaterial( instance, index, material ) );
}

void createLightMaterial( RTcontext context, RTgeometryinstance instance, int index, OBJREC* rec )
{
	RTmaterial material;
	RTprogram closest_hit_program;
	//RTprogram any_hit_program;

	/* Create our hit programs to be shared among all materials */
	sprintf( path_to_ptx, "%s/%s", sutilSamplesPtxDir(), "cuda_compile_ptx_generated_radiance_light.cu.ptx" );
	RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "closest_hit_radiance", &closest_hit_program ) );
	//RT_CHECK_ERROR( rtProgramCreateFromPTXFile( context, path_to_ptx, "any_hit_shadow", &any_hit_program ) );

	RT_CHECK_ERROR( rtMaterialCreate( context, &material ) );

	/* Note that we are leaving anyHitProgram[0] and closestHitProgram[1] as NULL.
	 * This is because our radiance rays only need closest_hit and shadow rays only
	 * need any_hit */
	RT_CHECK_ERROR( rtMaterialSetClosestHitProgram( material, RADIANCE_RAY, closest_hit_program ) );
	//RT_CHECK_ERROR( rtMaterialSetAnyHitProgram( material, SHADOW_RAY, any_hit_program ) );

	/* Set variables to be consumed by material for this geometry instance */
	applyMaterialVariable3f( context, material, "color", rec->oargs.farg[0], rec->oargs.farg[1], rec->oargs.farg[2] );

	/* Apply this material to the geometry instance. */
	RT_CHECK_ERROR( rtGeometryInstanceSetMaterial( instance, index, material ) );
}

void createAcceleration( RTcontext context, RTgeometryinstance instance )
{
	RTgeometrygroup geometrygroup;
	RTacceleration  acceleration;

	/* Create a geometry group to hold the geometry instance.  This will be used as the top level group. */
	RT_CHECK_ERROR( rtGeometryGroupCreate( context, &geometrygroup ) );
	RT_CHECK_ERROR( rtGeometryGroupSetChildCount( geometrygroup, 1 ) );
	RT_CHECK_ERROR( rtGeometryGroupSetChild( geometrygroup, 0, instance ) );

	/* Set the geometry group as the top level object. */
	applyContextObject( context, "top_object", geometrygroup );
	applyContextObject( context, "top_shadower", geometrygroup );
	if ( calc_ambient ^ !use_ambient )
		applyContextObject( context, "top_ambient", geometrygroup ); // TODO It would be nice not to need to do this

	/* create acceleration object for group and specify some build hints */
	RT_CHECK_ERROR( rtAccelerationCreate( context, &acceleration ) );
	RT_CHECK_ERROR( rtAccelerationSetBuilder( acceleration, "Sbvh" ) );
	RT_CHECK_ERROR( rtAccelerationSetTraverser( acceleration, "Bvh" ) );
	RT_CHECK_ERROR( rtAccelerationSetProperty( acceleration, "vertex_buffer_name", "vertex_buffer" ) ); // For Sbvh only
	RT_CHECK_ERROR( rtAccelerationSetProperty( acceleration, "index_buffer_name", "vindex_buffer" ) ); // For Sbvh only
	RT_CHECK_ERROR( rtGeometryGroupSetAcceleration( geometrygroup, acceleration ) );

	/* mark acceleration as dirty */
	RT_CHECK_ERROR( rtAccelerationMarkDirty( acceleration ) );
}

/* Helper functions */

void applyContextVariable1ui( RTcontext context, char* name, unsigned int value )
{
	RTvariable var;
	RT_CHECK_ERROR( rtContextDeclareVariable( context, name, &var ) );
	RT_CHECK_ERROR( rtVariableSet1ui( var, value ) );
}

void applyContextVariable1i( RTcontext context, char* name, int value )
{
	RTvariable var;
	RT_CHECK_ERROR( rtContextDeclareVariable( context, name, &var ) );
	RT_CHECK_ERROR( rtVariableSet1i( var, value ) );
}

void applyContextVariable1f( RTcontext context, char* name, float value )
{
	RTvariable var;
	RT_CHECK_ERROR( rtContextDeclareVariable( context, name, &var ) );
	RT_CHECK_ERROR( rtVariableSet1f( var, value ) );
}

void applyContextVariable3f( RTcontext context, char* name, float x, float y, float z )
{
	RTvariable var;
	RT_CHECK_ERROR( rtContextDeclareVariable( context, name, &var ) );
	RT_CHECK_ERROR( rtVariableSet3f( var, x, y, z ) );
}

void applyProgramVariable1ui( RTcontext context, RTprogram program, char* name, unsigned int value )
{
	RTvariable var;
	RT_CHECK_ERROR( rtProgramDeclareVariable( program, name, &var ) );
	RT_CHECK_ERROR( rtVariableSet1ui( var, value ) );
}

void applyProgramVariable1f( RTcontext context, RTprogram program, char* name, float value )
{
	RTvariable var;
	RT_CHECK_ERROR( rtProgramDeclareVariable( program, name, &var ) );
	RT_CHECK_ERROR( rtVariableSet1f( var, value ) );
}

void applyProgramVariable2f( RTcontext context, RTprogram program, char* name, float x, float y )
{
	RTvariable var;
	RT_CHECK_ERROR( rtProgramDeclareVariable( program, name, &var ) );
	RT_CHECK_ERROR( rtVariableSet2f( var, x, y ) );
}

void applyProgramVariable3f( RTcontext context, RTprogram program, char* name, float x, float y, float z )
{
	RTvariable var;
	RT_CHECK_ERROR( rtProgramDeclareVariable( program, name, &var ) );
	RT_CHECK_ERROR( rtVariableSet3f( var, x, y, z ) );
}

void applyMaterialVariable1ui( RTcontext context, RTmaterial material, char* name, unsigned int value )
{
	RTvariable var;
	RT_CHECK_ERROR( rtMaterialDeclareVariable( material, name, &var ) );
	RT_CHECK_ERROR( rtVariableSet1ui( var, value ) );
}

void applyMaterialVariable1f( RTcontext context, RTmaterial material, char* name, float value )
{
	RTvariable var;
	RT_CHECK_ERROR( rtMaterialDeclareVariable( material, name, &var ) );
	RT_CHECK_ERROR( rtVariableSet1f( var, value ) );
}

void applyMaterialVariable3f( RTcontext context, RTmaterial material, char* name, float x, float y, float z )
{
	RTvariable var;
	RT_CHECK_ERROR( rtMaterialDeclareVariable( material, name, &var ) );
	RT_CHECK_ERROR( rtVariableSet3f( var, x, y, z ) );
}

void createBuffer1D( RTcontext context, RTbuffertype type, RTformat format, int element_count, RTbuffer* buffer )
{
	RT_CHECK_ERROR( rtBufferCreate( context, type, buffer ) );
	RT_CHECK_ERROR( rtBufferSetFormat( *buffer, format ) );
	RT_CHECK_ERROR( rtBufferSetSize1D( *buffer, element_count ) );
}

void createCustomBuffer1D( RTcontext context, RTbuffertype type, int element_size, int element_count, RTbuffer* buffer )
{
	RT_CHECK_ERROR( rtBufferCreate( context, type, buffer ) );
	RT_CHECK_ERROR( rtBufferSetFormat( *buffer, RT_FORMAT_USER ) );
	RT_CHECK_ERROR( rtBufferSetElementSize( *buffer, element_size ) );
	RT_CHECK_ERROR( rtBufferSetSize1D( *buffer, element_count ) );
}

void createBuffer2D( RTcontext context, RTbuffertype type, RTformat format, int x_count, int y_count, RTbuffer* buffer )
{
	RT_CHECK_ERROR( rtBufferCreate( context, type, buffer ) );
	RT_CHECK_ERROR( rtBufferSetFormat( *buffer, format ) );
	RT_CHECK_ERROR( rtBufferSetSize2D( *buffer, x_count, y_count ) );
}

void createCustomBuffer2D( RTcontext context, RTbuffertype type, int element_size, int x_count, int y_count, RTbuffer* buffer )
{
	RT_CHECK_ERROR( rtBufferCreate( context, type, buffer ) );
	RT_CHECK_ERROR( rtBufferSetFormat( *buffer, RT_FORMAT_USER ) );
	RT_CHECK_ERROR( rtBufferSetElementSize( *buffer, element_size ) );
	RT_CHECK_ERROR( rtBufferSetSize2D( *buffer, x_count, y_count ) );
}

void createBuffer3D( RTcontext context, RTbuffertype type, RTformat format, int x_count, int y_count, int z_count, RTbuffer* buffer )
{
	RT_CHECK_ERROR( rtBufferCreate( context, type, buffer ) );
	RT_CHECK_ERROR( rtBufferSetFormat( *buffer, format ) );
	RT_CHECK_ERROR( rtBufferSetSize3D( *buffer, x_count, y_count, z_count ) );
}

void createCustomBuffer3D( RTcontext context, RTbuffertype type, int element_size, int x_count, int y_count, int z_count, RTbuffer* buffer )
{
	RT_CHECK_ERROR( rtBufferCreate( context, type, buffer ) );
	RT_CHECK_ERROR( rtBufferSetFormat( *buffer, RT_FORMAT_USER ) );
	RT_CHECK_ERROR( rtBufferSetElementSize( *buffer, element_size ) );
	RT_CHECK_ERROR( rtBufferSetSize2D( *buffer, x_count, y_count ) );
	RT_CHECK_ERROR( rtBufferSetSize3D( *buffer, x_count, y_count, z_count ) );
}

void applyContextObject( RTcontext context, char* name, RTobject object )
{
	RTvariable var;
	RT_CHECK_ERROR( rtContextDeclareVariable( context, name, &var ) );
	RT_CHECK_ERROR( rtVariableSetObject( var, object ) );
}

void applyGeometryObject( RTcontext context, RTgeometry geometry, char* name, RTobject object )
{
	RTvariable var;
	RT_CHECK_ERROR( rtGeometryDeclareVariable( geometry, name, &var ) );
	RT_CHECK_ERROR( rtVariableSetObject( var, object ) );
}

#ifdef TIMEOUT_CALLBACK
// Enable the use of OptiX timeout callbacks to help reduce the likelihood
// of kernel timeouts when rendering on GPUs that are also used for display
static int timeoutCallback(void)
{
	int earlyexit = 0;
	fprintf(stderr, "OptiX kernel running...\r\n");
	return earlyexit; 
}
#endif

//#endif /* OPTIX */